#include <qdrp/qdrp.h>
#include "../wpan/p802_15_4nam.h"
// **************************************************************************************************************************************************
// ENLACES CON LOS SCRIPTS TCL
// **************************************************************************************************************************************************

/* Enlace de la cabecera QDRP */
int hdr_qdrp::offset_;
static class QDRPHeaderClass : public PacketHeaderClass
{
public: 
	QDRPHeaderClass() : PacketHeaderClass("PacketHeader/QDRP", QDRP_HEADER_LENGTH)
	{
		bind_offset(&hdr_qdrp::offset_);
	}
} class_rtProtoQDRP_header;


/* Enlace del Agente QDRP */
static class QDRPclass : public TclClass
{
public:
	QDRPclass() : TclClass("Agent/QDRP") {}
	TclObject* create(int argc, const char*const* argv)
	{
		assert(argc==5);
		return (new QDRP((nsaddr_t)Address::instance().str2addr(argv[4])));
	}
} class_rtProtoQDRP;

// **************************************************************************************************************************************************
// FIN DE: ENLACES CON LOS SCRIPTS TCL
// **************************************************************************************************************************************************





// **************************************************************************************************************************************************
// TEMPORIZADORES
// **************************************************************************************************************************************************

// Temporizador para unirse a una red
void qdrp_timer__join_network::expire(Event* e){ agent_->current_state()->On_timer__join_network(); }

// Temporizador periódico para la emisión de SHOW_OFFs 
void qdrp_timer__heartbeat::expire(Event* e){ agent_->current_state()->On_timer__heartbeat(); }

// Conexión con nodo padre
void qdrp_timer__parent_node_connection::expire(Event* e){ agent_->current_state()->On_timer__connection_with_parent_node(); }

/* Conexión con un nodo hijo */
void qdrp_timer__child_node_connection::expire(Event* e){ agent_->current_state()->On_timer__connection_with_child_node(child_id_); }

void qdrp_timer__sendme_show_off::expire(Event* e){ agent_->current_state()->On_timer__sendme_show_off(); }

void qdrp_timer__new_route::expire(Event* e){ agent_->current_state()->On_timer__new_route(); }

// **************************************************************************************************************************************************
// FIN DE: TEMPORIZADORES
// **************************************************************************************************************************************************










// **************************************************************************************************************************************************
// ESTADO DE UN NODO																
// **************************************************************************************************************************************************

//==========================================================================================================================//
// Constructor/Destructor 											     	    //
//==========================================================================================================================//
// Constructor //
qdrp_state* qdrp_state::create_state(QDRP* agent, u_int8_t state_type)
{
	switch(state_type)
	{
		case OFF:
			return new qdrp_state__off(agent);
			break;
		case BOOTING:
			return new qdrp_state__booting(agent);
			break;
		case OFFLINE:
			return new qdrp_state__offline(agent);
			break;
		case CONNECTING:
			return new qdrp_state__connecting(agent);
			break;
		case ONLINE:
			return new qdrp_state__online(agent);
			break;
		case DISCONNECTING:
			return new qdrp_state__disconnecting(agent);
			break;
		default:
			fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: attempt to create a QDRP state object (state_type: %d) which does not exist\n", QDRP_global::CURRENT_TIME(), agent->this_node()->address(), state_type);
			exit(-1);
			return NULL;
			break;
	}
}

// Destructor //
qdrp_state::~qdrp_state(){};

// **************************************************************************************************************************************************
// FIN DE: ESTADO DE UN NODO
// **************************************************************************************************************************************************



// =============================================================================================================================================== //
// ESTADO = OFF																	   //
// =============================================================================================================================================== //
// ======================================================================================================================== //
// Constructor/Destructor 											     	    //
// ======================================================================================================================== //
// Constructor //
qdrp_state__off::qdrp_state__off(QDRP* agent){ agent_=agent; }

// Destructor //
qdrp_state__off::~qdrp_state__off(){}

// ======================================================================================================================== //
// Inicialización de estado											     	    //
// ======================================================================================================================== //
void qdrp_state__off::initialize()
{
	fprintf(stdout,"[QDRP] at time %f, node %d is OFF\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address());

}

// ======================================================================================================================== //
// Temporizadores												     	    //
// ======================================================================================================================== //
void qdrp_state__off::On_timer__join_network()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"join_network\" timer signal while OFF\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}
void qdrp_state__off::On_timer__heartbeat()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"heartbeat\" timer signal while OFF\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}
void qdrp_state__off::On_timer__connection_with_parent_node()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"connection_with_parent_node\" timer signal while OFF\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}
void qdrp_state__off::On_timer__connection_with_child_node(u_int8_t child_id)
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_child_node\" timer signal while OFF\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}	

void qdrp_state__off::On_timer__sendme_show_off()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"sendme_showoff\" timer signal while OFF\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__off::On_timer__new_route()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"new_route\" timer signal while OFF\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

// ======================================================================================================================== //
// Comandos		 											     	    //
// ======================================================================================================================== //
void qdrp_state__off::On_command__boot()
{
	agent_->node_is_sink()=false;
	agent_->network_id()=QDRP_global::DEFAULT_NETWORK_ID;
	agent_->change_state(qdrp_state::BOOTING);	
}

void qdrp_state__off::On_command__boot_sink(u_int8_t network_id)
{
	agent_->node_is_sink()=true;
	agent_->network_id()=network_id;
	agent_->change_state(qdrp_state::BOOTING);
}
void qdrp_state__off::On_command__switch_off()
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already OFF. \"switch-off\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
}

// ======================================================================================================================== //
// Recepción de paquetes											     	    //
// ======================================================================================================================== //	
// Por ahora no conozco ningún dispositivo capaz de transmitir y/o recibir mensajes estado apagado :P .
void qdrp_state__off::On_reception_of_data_packet(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__show_off(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__connect(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__connection_accepted(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__connection_rejected(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__connection_renew(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__disconnect(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__disconnection_ok(Packet* p){ Packet::free(p); }
void qdrp_state__off::On_reception_of_qdrp_packet__route_crash(Packet* p){ Packet::free(p); }

void qdrp_state__off::On_reception_of_qdrp_packet__sendme_show_off(Packet* p){ Packet::free(p); }
// =============================================================================================================================================== //
// FIN DE: ESTADO = OFF																   //
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = ARRANCANDO																   //
// =============================================================================================================================================== //

//==========================================================================================================================//
// Constructor/Destructor 											     	    //
//==========================================================================================================================//
qdrp_state__booting::qdrp_state__booting(QDRP* agent){ agent_=agent; }		// Constructor //

qdrp_state__booting::~qdrp_state__booting(){}					// Destructor //

//==========================================================================================================================//
// Inicialización de estado 											     	    //
//==========================================================================================================================//
void qdrp_state__booting::initialize()
{
	fprintf(stdout, "[QDRP] at time %f, node %d is booting\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address());


	// Limpia la lista de nodos descubiertos
	agent_->discovered_node_list()->clear();

	// Borra (si ya existía) el mapa de direcciones de nodos hijos existente
	if(agent_->child_node_address_map()!=NULL)
	{
		delete agent_->child_node_address_map();
		agent_->child_node_address_map()=NULL;
	}

	// El nodo padre no se conoce aún (o bien, si el nodo es un sumidero, no tiene nodo padre)
	agent_->parent_node_address() = (nsaddr_t) IP_BROADCAST;

	if (agent_->node_is_sink())
	{
		// Crea la tabla de direcciones de nodos hijos
		agent_->child_node_address_map() = new qdrp_address_map(QDRP_global::SINK__NUM_CHILD_SLOTS);
		
		// Establece los parámetros de un nodo sumidero
		agent_->this_node()->hops_to_sink() = 0;
		agent_->this_node()->network_id() = agent_->network_id();
		agent_->this_node()->num_child_nodes() = 0;
		agent_->this_node()->max_child_nodes() = QDRP_global::SINK__NUM_CHILD_SLOTS;
		
		// Cambia el estado actual a CONECTADO (consigo mismo :D )
		agent_->change_state(qdrp_state::ONLINE);
	}
	else
	{
		// Crea el mapa de direcciones de nodos hijos
		agent_->child_node_address_map() = new qdrp_address_map(QDRP_global::NODE__NUM_CHILD_SLOTS);

		// Establece los parámetros de un nodo normal
		agent_->this_node()->hops_to_sink() = QDRP_global::UNKNOWN_HOPS_TO_SINK;
		agent_->this_node()->network_id() = agent_->network_id();
		agent_->this_node()->num_child_nodes() = 0;
		agent_->this_node()->max_child_nodes() = QDRP_global::NODE__NUM_CHILD_SLOTS;

		// Cambia el estado actual a DESCONECTADO
		agent_->change_state(qdrp_state::OFFLINE);
	}	
}

//==========================================================================================================================//
// Temporizadores												     	    //
//==========================================================================================================================//
void qdrp_state__booting::On_timer__join_network()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"join_network\" timer signal while booting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__booting::On_timer__heartbeat()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"heartbeat\" timer signal while booting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__booting::On_timer__connection_with_parent_node()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_parent_node\" timer signal while booting\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__booting::On_timer__connection_with_child_node(u_int8_t child_id)
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_child_node\" timer signal while booting\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__booting::On_timer__sendme_show_off()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"sendme_showoff\" timer signal while OFF\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__booting::On_timer__new_route()
{
	fprintf(stdout,"[QDRP] at time %f, node %d :: ERROR: received \"new_route\" timer signal while booting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}


//==========================================================================================================================//
// Comandos													     	    //
//==========================================================================================================================//
void qdrp_state__booting::On_command__boot()
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}

void qdrp_state__booting::On_command__boot_sink(u_int8_t network_id)
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot-sink\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
}

void qdrp_state__booting::On_command__switch_off()
{
	agent_->switch_off();
}

// ======================================================================================================================== //
// Recepción de paquetes											     	    //
// ======================================================================================================================== //	
// En estado "arrancando" no hay TXon ni RXon de mensajes en el nodo
void qdrp_state__booting::On_reception_of_data_packet(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__show_off(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__connect(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__connection_accepted(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__connection_rejected(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__connection_renew(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__disconnect(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__disconnection_ok(Packet* p){ Packet::free(p); }
void qdrp_state__booting::On_reception_of_qdrp_packet__route_crash(Packet* p){ Packet::free(p); }

void qdrp_state__booting::On_reception_of_qdrp_packet__sendme_show_off(Packet* p){ Packet::free(p); }
// =============================================================================================================================================== //
// FIN DE: ESTADO = ARRANCANDO															   //
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = DESCONECTADO (IMPLEMENTACIÓN)													   //
// =============================================================================================================================================== //
// ======================================================================================================================== //
// Constructor													     	    //
// ======================================================================================================================== //
qdrp_state__offline::qdrp_state__offline(QDRP* agent){ agent_=agent; } 		// Constructor //

qdrp_state__offline::~qdrp_state__offline(){}					// Destructor //

// ======================================================================================================================== //
// Inicialización de estado											     	    //
// ======================================================================================================================== //
void qdrp_state__offline::initialize()
{
	fprintf(stdout, "[QDRP] at time %f, node %d is offline\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address());

//double rand=Random::normal(0.05, 0.80);
Tcl& tcl = Tcl::instance();

tcl.evalf("[Simulator instance] at %.9f {[Simulator instance] puts-nam-traceall {n -t %.9f -s %d -S DLABEL -l [offline] -L \"\"}}",QDRP_global::CURRENT_TIME()+0.001,QDRP_global::CURRENT_TIME(),agent_->this_node()->address());


	agent_->sendme_show_off_timer()->resched(14.0);
	agent_->join_network_timer()->resched(16.1);

}

// ======================================================================================================================== //
// Temporizadores												     	    //
// ======================================================================================================================== //
void qdrp_state__offline::On_timer__join_network()
{
Tcl& tcl = Tcl::instance();

QDRP_global::cancel_timer(agent_->sendme_show_off_timer());


	// si la lista de nodos descubiertos está vacía, no se puede hacer nada aún...
	if(agent_->discovered_node_list()->is_empty())
	{
		// ...salvo recargar el temporizador :P

		agent_->sendme_show_off_timer()->resched(QDRP_global::sendme_show_off_interval());
		agent_->join_network_timer()->resched(QDRP_global::JOIN_NETWORK_INTERVAL());


		return;
	}

	// si hay algún nodo al que conectarse...
	// cambia el estado a CONECTANDO



	agent_->change_state(qdrp_state::CONNECTING);	
}


void qdrp_state__offline::On_timer__heartbeat()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"heartbeat\" timer signal while offline\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);	
}


void qdrp_state__offline::On_timer__connection_with_parent_node()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_parent_node\" timer signal while offline\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);	
}


void qdrp_state__offline::On_timer__connection_with_child_node(u_int8_t child_id)
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_child_node\" timer signal while offline\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		
}

void qdrp_state__offline::On_timer__sendme_show_off() //xendme
{

Tcl& tcl = Tcl::instance();

char note[80] = "";

sprintf(note, "OFF LINE: \tel nodo %d envio un SENDME_SHOW_OFF en el tiempo\t%.9f", agent_->this_node()->address(), QDRP_global::CURRENT_TIME());

tcl.evalf("[Simulator instance] at %.9f {[Simulator instance] trace-annotate \"%s\"}",QDRP_global::CURRENT_TIME()+0.001,note);

agent_->sendme_show_off_timer()->resched(QDRP_global::sendme_show_off_interval());


agent_->send_qdrp_packet__sendme_show_off();



}


void qdrp_state__offline::On_timer__new_route()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"new_route\" timer signal while offline\n", 			QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		
}


// ======================================================================================================================== //
// Comandos													     	    //
// ======================================================================================================================== //
void qdrp_state__offline::On_command__boot()
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}


void qdrp_state__offline::On_command__boot_sink(u_int8_t network_id)
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot-sink\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
}


void qdrp_state__offline::On_command__switch_off()
{
	agent_->switch_off();
}



// ======================================================================================================================== //
// Recepción de paquetes											     	    //
// ======================================================================================================================== //	
void qdrp_state__offline::On_reception_of_data_packet(Packet* p)
{
	// Como este nodo no tiene ruta hacia un sumidero, se descarta el paquete
	agent_->drop(p, DROP_RTR_NO_ROUTE); 
}

void qdrp_state__offline::On_reception_of_qdrp_packet__show_off(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);
	struct hdr_qdrp_packet__show_off* qdrp_header = HDR_QDRP_PACKET__SHOW_OFF(p);

FILE *fichero;

char note[100] = "";
double rand;


 
//si recibo un show off del sumidero
if(agent_->discovered_node_list()->is_empty() && qdrp_header->hops_to_sink() == 0)
{

QDRP_global::cancel_timer(agent_->join_network_timer());
QDRP_global::cancel_timer(agent_->sendme_show_off_timer());

   rand=Random::uniform(0.10, 0.45);
   agent_->join_network_timer()->resched((rand<0)?(rand*-1):rand);


}

else
if(agent_->discovered_node_list()->is_empty())
{

	agent_->join_network_timer()->resched(QDRP_global::JOIN_NETWORK_INTERVAL());
	agent_->sendme_show_off_timer()->resched(QDRP_global::sendme_show_off_interval());




}


	// Construye la información del nodo
	qdrp_node source_node;
	source_node.address()=ip_header->saddr();
	source_node.network_id()=qdrp_header->network_id();
	source_node.hops_to_sink()=qdrp_header->hops_to_sink();
	source_node.num_child_nodes()=qdrp_header->num_child_nodes();
	source_node.max_child_nodes()=qdrp_header->max_child_nodes();
	source_node.lqi_packet()=ip_header->lqi_packet;


	// Actualiza la lista de nodos conocidos
	agent_->discovered_node_list()->register_node(source_node);


	// Libera los recursos asociados al paquete
	Packet::free(p);
}



void qdrp_state__offline::On_reception_of_qdrp_packet__connect(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__connection_accepted(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__connection_rejected(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__connection_renew(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__disconnect(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__disconnection_ok(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__route_crash(Packet* p){ Packet::free(p); }
void qdrp_state__offline::On_reception_of_qdrp_packet__sendme_show_off(Packet* p){ Packet::free(p); }	//xendme
// =============================================================================================================================================== //
// FIN DE: ESTADO = DESCONECTADO														   //
// =============================================================================================================================================== //




// =============================================================================================================================================== //
// ESTADO = CONECTANDO																   //
// =============================================================================================================================================== //
// ======================================================================================================================== //
// Constructor/Destructor 											     	    //
// ======================================================================================================================== //
qdrp_state__connecting::qdrp_state__connecting(QDRP* agent){ agent_=agent; }	// Constructor //

qdrp_state__connecting::~qdrp_state__connecting(){}	// Destructor //


// ======================================================================================================================== //
// Inicialización de estado											     	    //
// ======================================================================================================================== //
void qdrp_state__connecting::initialize()
{	
	// si no hay ningún nodo más al que conectarse, volvemos al estado "fuera de línea"
	if(agent_->discovered_node_list()->is_empty())
	{
		agent_->send_qdrp_packet__sendme_show_off();
		agent_->change_state(qdrp_state::OFFLINE);
		return;
	}

double rand;

	QDRP_global::cancel_timer(agent_->sendme_show_off_timer());
	QDRP_global::cancel_timer(agent_->new_route_timer());

	// establece la dirección del nodo candidato como nodo padre
	agent_->parent_node_address() = agent_->discovered_node_list()->get_best_candidate().address();



	fprintf(stdout, "[QDRP] at time %f, node %d is connecting to node %d\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address(), agent_->parent_node_address());

	// envía "conectar" al candidato a nodo padre
	agent_->send_qdrp_packet__connect(agent_->parent_node_address());

rand=Random::normal(0.05, 0.50);

	// y activa un temporizador de conexión
	agent_->parent_node_connection_timer()->resched(1.5+((rand<0)?(rand*-1):rand)/*QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL()*/);
	agent_->parent_node_connection_timer()->cycle()=0;
}


// ======================================================================================================================== //
// Temporizadores												     	    //
// ======================================================================================================================== //
void qdrp_state__connecting::On_timer__join_network()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"join_network\" timer signal while connecting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		
}


void qdrp_state__connecting::On_timer__heartbeat()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"heartbeat\" timer signal while connecting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		
}


void qdrp_state__connecting::On_timer__connection_with_parent_node()
{

double rand=Random::uniform(0.05, 0.50);

	switch(++agent_->parent_node_connection_timer()->cycle())
	{
		case 1:
			agent_->send_qdrp_packet__connect(agent_->parent_node_address()); 					
		 	agent_->parent_node_connection_timer()->resched(2.0+((rand<0)?(rand*-1):rand));
			break;

		case 2:
			agent_->send_qdrp_packet__connect(agent_->parent_node_address()); 
		 	agent_->parent_node_connection_timer()->resched(2.0+((rand<0)?(rand*-1):rand));
			break;

		case 3:
			agent_->send_qdrp_packet__connect(agent_->parent_node_address()); 
			agent_->parent_node_connection_timer()->resched(2.0+((rand<0)?(rand*-1):rand));
			break;

		case 4:
			agent_->send_qdrp_packet__connect(agent_->parent_node_address()); 
		 	agent_->parent_node_connection_timer()->resched(2.0+((rand<0)?(rand*-1):rand));
			break;


		case 5: // (n+1)ésimo reintento de conexión (el candidato a nodo padre no contesta)

			agent_->send_qdrp_packet__connect(agent_->parent_node_address()); 
			agent_->parent_node_connection_timer()->resched(2.0+((rand<0)?(rand*-1):rand));
			break;
		case 6: // El candidato a nodo padre ni se ha enterado de la solicitud, ha caído, o simplemente se ha hecho el sueco.
			On_connection_attempt_failure();
			break;
		default: 
			fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: parent_node_connection_timer cycle counter went out of range\n", 					QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
			exit(-1);
			break;
	}
}


void qdrp_state__connecting::On_timer__connection_with_child_node(u_int8_t child_id)
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_child_node\" timer signal while connecting\n", 		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		
}

void qdrp_state__connecting::On_timer__sendme_show_off()	
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR->: received \"sendme_show_off\" timer signal while connecting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		

}

void qdrp_state__connecting::On_timer__new_route()
{
	fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"new_route\" timer signal while connecting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);		
}

// ======================================================================================================================== //
// Comandos													     	    //
// ======================================================================================================================== //
void qdrp_state__connecting::On_command__boot()
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}

void qdrp_state__connecting::On_command__boot_sink(u_int8_t network_id)
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot-sink\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}

void qdrp_state__connecting::On_command__switch_off()
{
	agent_->switch_off();
}


// ======================================================================================================================== //
// Recepción de paquetes											     	    //
// ======================================================================================================================== //
void qdrp_state__connecting::On_reception_of_data_packet(Packet* p)
{
	// no estando aún conectado el nodo no puede enviar ni enrutar paquetes de datos.
	agent_->drop(p, DROP_RTR_NO_ROUTE); 
}


void qdrp_state__connecting::On_reception_of_qdrp_packet__show_off(Packet* p){ Packet::free(p); }
void qdrp_state__connecting::On_reception_of_qdrp_packet__connect(Packet* p){ Packet::free(p); }
void qdrp_state__connecting::On_reception_of_qdrp_packet__connection_renew(Packet* p){ Packet::free(p); }
void qdrp_state__connecting::On_reception_of_qdrp_packet__disconnect(Packet* p){ Packet::free(p); }

void qdrp_state__connecting::On_reception_of_qdrp_packet__disconnection_ok(Packet* p){ Packet::free(p); }


void qdrp_state__connecting::On_reception_of_qdrp_packet__connection_accepted(Packet* p)
{	
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header = HDR_IP(p);

	// Se asegura de que el mensaje proceda del candidato a padre Y sea para este nodo
	if((ip_header->saddr() != agent_->parent_node_address()) || (ip_header->daddr() != agent_->this_node()->address()))
	{
		Packet::free(p);
		return;
	}
	
//Diego UPV
	Tcl& tcl = Tcl::instance();

	float attime = QDRP_global::CURRENT_TIME();
	
	if (attime > 0.0)
		attime += 0.000100000;

	tcl.evalf("[Simulator instance] at %.9f {[Simulator instance] puts-nam-traceall {n -t %.9f -s %d -S DLABEL -l [p-%d] -L \"\"}}",attime,attime,ip_header->daddr(), ip_header->saddr());


	Packet::free(p); // Libera los recursos asociados al paquete	
	// OK, este mensaje es para nosotros. El que hasta ahora ha sido "candidato a" padre ya es padre oficialmente :D
	On_connection_attempt_success();
}


void qdrp_state__connecting::On_reception_of_qdrp_packet__connection_rejected(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header = HDR_IP(p);

	// Se asegura de que el mensaje proceda del candidato a padre y que sea para este nodo
	if((ip_header->saddr() != agent_->parent_node_address()) || (ip_header->daddr() != agent_->this_node()->address()))
	{
		Packet::free(p);
		return;
	}

	Packet::free(p); // Libera los recursos asociados al paquete (que ya no hace falta)

	// El candidato a nodo padre rechazó la conexión.
	On_connection_attempt_failure();
}


void qdrp_state__connecting::On_reception_of_qdrp_packet__route_crash(Packet* p)
{
	// El código es similar al de "On_reception_of_qdrp_packet__connection_rejected" salvo en que en éste no es necesario       //
	// comprobar que somos el destinatario del paquete, ya que ROUTE_CRASH se envía por difusión				    //

	// Obtiene la cabecera IP
	struct hdr_ip* ip_header = HDR_IP(p);

	// Comprueba que procede del candidato a padre
	if(ip_header->saddr() != agent_->parent_node_address()){ Packet::free(p); return; }

	Packet::free(p); // Libera los recursos asociados al paquete (que ya no hace falta)

	// Oops...el propio candidato a padre se ha quedado sin conexión. :?
	On_connection_attempt_failure();
}

void qdrp_state__connecting::On_reception_of_qdrp_packet__sendme_show_off(Packet* p){ Packet::free(p); }

// ======================================================================================================================== //
// Funciones internas												     	    //
// ======================================================================================================================== //
// acción tomada tras un intento de conexión fallido
void qdrp_state__connecting::On_connection_attempt_failure()
{
	// Detiene el temporizador de conexión
	QDRP_global::cancel_timer(agent_->parent_node_connection_timer());

	
	agent_->parent_node_connection_timer()->cycle()=0;

	// Borra el ex-candidato a nodo padre de la lista de nodos descubiertos
	agent_->discovered_node_list()->delete_best_candidate();

	// des-referencia el nodo padre
	agent_->parent_node_address() = (nsaddr_t) IP_BROADCAST;		

	// y prueba suerte de nuevo, reiniciando el estado CONECTANDO
	agent_->change_state(qdrp_state::CONNECTING);			
}


// acción tomada tras un intento de conexión exitoso
void qdrp_state__connecting::On_connection_attempt_success()
{
	// Renueva el temporizador de conexión con el padre
	agent_->parent_node_connection_timer()->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
	agent_->parent_node_connection_timer()->cycle()=0;

	// actualiza la información de este nodo basándonos en la información del padre...
	agent_->this_node()->network_id() = agent_->discovered_node_list()->get_best_candidate().network_id();
	agent_->this_node()->hops_to_sink() = (agent_->discovered_node_list()->get_best_candidate().hops_to_sink()+1);

	// Vacía la lista de nodos candidatos: por ahora ya no es necesaria.
	agent_->discovered_node_list()->clear();
		
	// y cambia a estado CONECTADO
	agent_->change_state(qdrp_state::ONLINE);	
}

// =============================================================================================================================================== //
// FIN DE: ESTADO = CONECTANDO															   //
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = CONECTADO																   //
// =============================================================================================================================================== //
// ======================================================================================================================== //
// Constructor / Destructor 											     	    //
// ======================================================================================================================== //
qdrp_state__online::qdrp_state__online(QDRP* agent){ agent_=agent; }	// Constructor //

qdrp_state__online::~qdrp_state__online(){}	// Destructor //


// ======================================================================================================================== //
// Inicialización de estado 											     	    //
// ======================================================================================================================== //
void qdrp_state__online::initialize()
{

	fprintf(stdout, "[QDRP] at time %f, node %d is online\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address());


   if(/* !agent_->node_is_sink() || */agent_->this_node()->hops_to_sink() > 1)
   {
      agent_->heartbeat_timer()->resched(QDRP_global::HEARTBEAT_INTERVAL); //45
      agent_->new_route_timer()->resched(QDRP_global::new_route_int); //70
   }

   if(agent_->node_is_sink())
      agent_->heartbeat_timer()->resched(7.0); //45


agent_->send_qdrp_packet__show_off();
}


// ======================================================================================================================== //
// Temporizadores												     	    //
// ======================================================================================================================== //
void qdrp_state__online::On_timer__join_network()
{
	fprintf(stdout, "[QDRP]: at time %f, node %d :: ERROR: received \"join_network\" timer signal while online\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);
}

void qdrp_state__online::On_timer__heartbeat()
{
//	 Recarga el temporizador "estoy vivo"

if(agent_->node_is_sink() && agent_->this_node()->num_child_nodes() == 0)
{
    agent_->heartbeat_timer()->resched(0.10); //45
    agent_->send_qdrp_packet__show_off();
}
	
if(!agent_->node_is_sink() || agent_->this_node()->hops_to_sink() > 1 )
{
	// envía un mensaje "show_off" siempre que pueda aceptar conexiones
		agent_->heartbeat_timer()->resched(QDRP_global::HEARTBEAT_INTERVAL); 
		agent_->send_qdrp_packet__sendme_show_off();

}

}


void qdrp_state__online::On_timer__connection_with_parent_node()
{
	switch(++agent_->parent_node_connection_timer()->cycle())
	{
		case 1:
		case 2:
		case 3:

		case 4:
		case 5: // en todos estos casos, no hace más que recargar el temporizador un tiempo de ciclo
			// (tiempo_de_ciclo = 1/6 tiempo_de_timeout)	
			agent_->parent_node_connection_timer()->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
			break;
		case 6: 
			// TIMEOUT => desastre: la conexión con el nodo padre se ha finiquitao...
			fprintf(stdout, "[QDRP] at time %f, node %d has timed-out connection with its parent node %d\n",
				QDRP_global::CURRENT_TIME(), agent_->this_node()->address(),agent_->parent_node_address());
			On_route_lost(true);
			break;
		default:
			fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: parent_node_connection_timer cycle counter went out of range\n",
				QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
			exit(-1);
			break;
	}
}


void qdrp_state__online::On_timer__connection_with_child_node(u_int8_t child_id)
{
	// busca la dirección del nodo hijo en el mapa de direcciones
	bool child_node_exists=false;
	nsaddr_t child_node_address=agent_->child_node_address_map()->get_address_by_identifier(child_id, &child_node_exists);

	if(!child_node_exists)
	{
		// Ha saltado un temporizador asociado a un nodo inexistente: es un BUG
		fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_child_node(%d)\" timer signal, associated with a node 				which does not exist\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address(), child_id);
		exit(-1);
	}

	switch(++agent_->child_node_connection_timer(child_id)->cycle())
	{
		case 1:
		case 2:
		case 3:
		case 4:
		case 5: // n-ésimo intento de renovación de conexión (a mayor n, mayor número de reconexiones que han quedado en agua de borrajas)
			agent_->send_qdrp_packet__connection_renew(child_node_address);
			agent_->child_node_connection_timer(child_id)->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
			break;
		case 6: // TIMEOUT => la conexión con el nodo hijo ha caído
			fprintf(stdout, "[QDRP] at time %f, node %d has timed-out connection with its child node %d\n", 
				QDRP_global::CURRENT_TIME(), agent_->this_node()->address(), child_node_address);	
			// restablece el contador de ciclos del temporizador de conexión con el hijo
			agent_->child_node_connection_timer(child_id)->cycle()=0;
			// desregistra el nodo hijo correspondiente
			agent_->child_node_address_map()->delete_entry(child_node_address);
			// y actualiza el nº de hijos del nodo
			agent_->this_node()->num_child_nodes()--;
			break;
		default:
			fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: parent_node_connection_timer cycle counter went out of range\n", 
				QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
			exit(-1);
			break;
	}
}

void qdrp_state__online::On_timer__sendme_show_off()
{
	fprintf(stdout, "[QDRP]: at time %f, node %d :: INFO: received \"sendme_show_off\" timer signal while online\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
//	exit(-1);	
}


void qdrp_state__online::On_timer__new_route()
{

int new_address;
qdrp_node new_parent;

QDRP_global::cancel_timer(agent_->heartbeat_timer());

   if(agent_->discovered_node_list()->is_empty())
   {
	agent_->heartbeat_timer()->resched(QDRP_global::HEARTBEAT_INTERVAL); //50
	agent_->new_route_timer()->resched(QDRP_global::new_route_int); //50
	return;
   }


   new_parent = agent_->discovered_node_list()->get_best_candidate();
   new_address = new_parent.address();

//si la nueva direccion es la misma del padre, no haga nada, retorne
   if(agent_->parent_node_address() == new_address)
{
	fprintf(stdout, "*******[QDRP]: at time %f, node %d :: ERROR: received \"MISMO_PADRE\" \n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());

      return;
}

//si el nuevo padre esta mas lejos que mi distancia actual, entonces nada, retorne
   if((new_parent.hops_to_sink()+1) >= agent_->this_node()->hops_to_sink())
   {
	fprintf(stdout, "*******[QDRP]: at time %f, node %d :: INFORMATION: received \"***MAYOR DIST > ****\" esta mas lejos\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
      return;
   }


	fprintf(stdout, "/////////////////////el padre de %d es \"%d\" y esta a %d saltos, Nuevo padre %d - a--------------- %d saltos\n",
		agent_->this_node()->address(), agent_->parent_node_address(), agent_->this_node()->hops_to_sink()-1, new_parent.address(), new_parent.hops_to_sink());


	agent_->heartbeat_timer()->resched(QDRP_global::HEARTBEAT_INTERVAL); //50
	agent_->new_route_timer()->resched(QDRP_global::new_route_int); //50


   agent_->send_qdrp_packet__connect(new_address);
//   new_parent->parent =

//	exit(-1);	


}

// ======================================================================================================================== //
// Comandos													     	    //
// ======================================================================================================================== //
void qdrp_state__online::On_command__boot()
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}

void qdrp_state__online::On_command__boot_sink(u_int8_t network_id)
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot-sink\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}

void qdrp_state__online::On_command__switch_off()
{
	agent_->switch_off();
}


// ======================================================================================================================== //
// Recepción de paquetes											     	    //
// ======================================================================================================================== //
// recepción de un paquete de datos //
void qdrp_state__online::On_reception_of_data_packet(Packet* p)
{
	struct hdr_cmn* common_header=HDR_CMN(p);
	struct hdr_ip* ip_header=HDR_IP(p);

	// Sentido del paquete ascendente (desde el nivel de enlace)
	if (common_header->direction() == hdr_cmn::UP && agent_->node_is_sink())
	{
		// Los sumideros no enrutan paquetes de datos, los absorben
		ip_header->daddr() = agent_->this_node()->address(); // toma "prestado" este paquete... 
		agent_->demultiplexer_->recv(p, 0);
		return;
	}			
	// Sentido del paquete descendente (dirigido AL nivel de enlace)
	else
	{
		common_header->addr_type()=NS_AF_INET;
		common_header->direction()=hdr_cmn::DOWN;

		if(ip_header->daddr() == ((nsaddr_t) IP_BROADCAST)) { common_header->next_hop() = IP_BROADCAST; }
	 	else
		{
			// como los paquetes de datos van únicamente hacia un sumidero, se limita a reenviar el paquete de datos al nodo padre
			common_header->next_hop() = agent_->parent_node_address(); 
			if(common_header->next_hop() == ((nsaddr_t) IP_BROADCAST))
			{
				agent_->drop(p, DROP_RTR_NO_ROUTE);	// Si no hay ruta, a la basura el paquete va
				return;
			}
		}
		agent_->send_data_packet(p);
	}
}


// recepción de paquetes QDRP //
void qdrp_state__online::On_reception_of_qdrp_packet__disconnection_ok(Packet* p){ Packet::free(p); }

void qdrp_state__online::On_reception_of_qdrp_packet__show_off(Packet* p){
// Packet::free(p);

	struct hdr_ip* ip_header = HDR_IP(p);
	struct hdr_qdrp_packet__show_off* qdrp_header = HDR_QDRP_PACKET__SHOW_OFF(p);


	if(agent_->node_is_sink() || ip_header->saddr() == ip_header->daddr())
	{
	   Packet::free(p);
	   return;
	}

	//si recibimos un show_off del padre, actualizamos nuestro numero de saltos
	if(ip_header->saddr() == agent_->parent_node_address())
	{
		agent_->this_node()->hops_to_sink() = qdrp_header->hops_to_sink()+1;
		agent_->send_qdrp_packet__show_off();
	}

//{
	// Construye la información del nodo
	qdrp_node source_node;
	source_node.address()		=	ip_header->saddr();
	source_node.network_id()	=	qdrp_header->network_id();
	source_node.hops_to_sink()	=	qdrp_header->hops_to_sink();
	source_node.num_child_nodes()	=	qdrp_header->num_child_nodes();
	source_node.max_child_nodes()	=	qdrp_header->max_child_nodes();
	source_node.lqi_packet()	=	ip_header->lqi_packet;

	// Actualiza la lista de nodos conocidos
	agent_->discovered_node_list()->register_node(source_node);
//}

	// Libera los recursos asociados al paquete

	Packet::free(p);
}


void qdrp_state__online::On_reception_of_qdrp_packet__connect(Packet* p)
{

	struct hdr_ip* ip_header=HDR_IP(p);


	// se asegura de que el mensaje vaya dirigido a este nodo
	if(ip_header->daddr()!=agent_->this_node()->address()){ Packet::free(p); return; }

	// comprueba si el paquete procede de un nodo hijo
	bool child_exists = true;
	u_int8_t child_id = agent_->child_node_address_map()->get_identifier_by_address(ip_header->saddr(), &child_exists);


	if(child_exists)
	{
		// la solicitud procede de un nodo hijo -> renovarle el temporizador
		agent_->child_node_connection_timer(child_id)->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
		agent_->child_node_connection_timer(child_id)->cycle()=0;
		
		// y envía "conexión aceptada"
		agent_->send_qdrp_packet__connection_accepted(ip_header->saddr());
	}
	else
	{
		// la solicitud no procede de un nodo hijo: comprueba si se pueden aceptar más conexiones
		if(agent_->child_node_address_map()->is_full() || agent_->this_node()->hops_to_sink() == QDRP_global::MAX_HOPS_TO_SINK)
		{
			// no se pueden aceptar más conexiones, rechaza el intento de conexión
			agent_->send_qdrp_packet__connection_rejected(ip_header->saddr());
		}
		else
		{
			// se pueden aceptar conexiones, así que registra el nodo solicitante
			agent_->child_node_address_map()->add_entry(ip_header->saddr());
			child_id = agent_->child_node_address_map()->get_identifier_by_address(ip_header->saddr(), &child_exists);
			if(!child_exists)
			{
				// imposible
				fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: could not register node %d\n",
					QDRP_global::CURRENT_TIME(), agent_->this_node()->address(), ip_header->saddr());
				exit(-1);
			}
			// actualiza el nº de hijos del nodo
			agent_->this_node()->num_child_nodes()++;
		
			// activa un temporizador asociado al nodo
			agent_->child_node_connection_timer(child_id)->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
			agent_->child_node_connection_timer(child_id)->cycle()=0;

			// y envía conexión aceptada al hijo
			agent_->send_qdrp_packet__connection_accepted(ip_header->saddr());
		}
	}
	Packet::free(p); // Libera los recursos asociados al paquete		
}


void qdrp_state__online::On_reception_of_qdrp_packet__connection_accepted(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);

	// Puede tratarse de una renovación de conexión. ¿Procede del nodo padre Y es para nosotros?
	if(ip_header->saddr() == agent_->parent_node_address() && ip_header->daddr() == agent_->this_node()->address())
	{
		assert(!agent_->node_is_sink());

		// Es una renovación de conexión :)
		agent_->parent_node_connection_timer()->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
		agent_->parent_node_connection_timer()->cycle()=0;
	}
	else
	if(ip_header->daddr() == agent_->this_node()->address())
	{
//envia desconectar al padre
		agent_->send_qdrp_packet__disconnect(agent_->parent_node_address());	

		agent_->parent_node_address() = ip_header->saddr(); //ojo con esto, mirar como traernos el valor de new_parent
		agent_->parent_node_connection_timer()->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
		agent_->parent_node_connection_timer()->cycle()=0;

	// actualiza la información de este nodo basándonos en la información del padre...
		agent_->this_node()->network_id() = agent_->discovered_node_list()->get_best_candidate().network_id();
		agent_->this_node()->hops_to_sink() = (agent_->discovered_node_list()->get_best_candidate().hops_to_sink()+1);

	// Vacía la lista de nodos candidatos: por ahora ya no es necesaria.
		agent_->discovered_node_list()->clear();
		
	}



	agent_->send_qdrp_packet__show_off();


	Packet::free(p); // Libera el paquete



}


void qdrp_state__online::On_reception_of_qdrp_packet__connection_rejected(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);

	// ¿Procede del nodo padre Y es para nosotros?
	if(ip_header->saddr() == agent_->parent_node_address() && ip_header->daddr() == agent_->this_node()->address())
	{
		assert(!agent_->node_is_sink());

		// Me pregunto que clase de mal funcionamiento podría causar este desaguisado. En teoría no debería suceder.
		On_route_lost(false);
	}
	Packet::free(p); // Libera el paquete
}


void qdrp_state__online::On_reception_of_qdrp_packet__connection_renew(Packet *p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);
	
	// ¿procede del nodo padre?
	if(ip_header->saddr()==agent_->parent_node_address() && !agent_->node_is_sink())
	{
		// es una renovación de conexión, a la que hay que responder con un CONNECT
		agent_->send_qdrp_packet__connect(ip_header->saddr());
		return;
	}
}


void qdrp_state__online::On_reception_of_qdrp_packet__disconnect(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);

	// el mensaje debe estar destinado a este nodo
	if(ip_header->daddr() != agent_->this_node()->address()){ Packet::free(p); return; }

	// ¿el mensaje procede de un nodo hijo?
	bool child_exists=false;
	u_int8_t child_id=agent_->child_node_address_map()->get_identifier_by_address(ip_header->saddr(), &child_exists);
		
	if(child_exists)
	{
		// OK, el mensaje procede de un nodo hijo. Desactiva el temporizador asociado
		QDRP_global::cancel_timer(agent_->child_node_connection_timer(child_id));
		agent_->child_node_connection_timer(child_id)->cycle()=0;

		// Envía DESCONEXIÓN_OK al hijo
		agent_->send_qdrp_packet__disconnection_ok(ip_header->saddr());
	
		// des-registra el nodo hijo
		agent_->child_node_address_map()->delete_entry(ip_header->saddr());

		// y actualiza el nº de hijos del nodo
		agent_->this_node()->num_child_nodes()--;
	}
	Packet::free(p);
}


void qdrp_state__online::On_reception_of_qdrp_packet__route_crash(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);

	// El mensaje sólo nos afecta si el nodo padre es el emisor
	if(ip_header->saddr()==agent_->parent_node_address())
	{
		assert(!agent_->node_is_sink());

		Packet::free(p);
		On_route_lost(false);
	}
	else Packet::free(p);
}

void qdrp_state__online::On_reception_of_qdrp_packet__sendme_show_off(Packet* p)
{
	
	// envía un mensaje "show_off" siempre que pueda aceptar conexiones
	if(!agent_->child_node_address_map()->is_full())
	{

		agent_->send_qdrp_packet__show_off();

	}
	else
	   Packet::free(p);
   
}


// ======================================================================================================================== //
// Funciones internas												     	    //
// ======================================================================================================================== //
// acción tomada cuando se pierde la conexión con el nodo padre // 
void qdrp_state__online::On_route_lost(bool timed_out)
{
	// Detiene el temporizador "estoy vivo"
	QDRP_global::cancel_timer(agent_->heartbeat_timer());

	if(timed_out)
	{
		// restablece el contador de ciclos
		agent_->parent_node_connection_timer()->cycle()=0; 
		
		// des-registra el nodo padre
		agent_->parent_node_address()=IP_BROADCAST;
	}
	else
	{
		// Envía DESCONECTAR al nodo padre para acelerar el proceso de desconexión...
		agent_->send_qdrp_packet__disconnect(agent_->parent_node_address());
	}

	// ¿hay nodos hijos?
	if(!agent_->child_node_address_map()->is_empty())
	{
		// si los hay, envía ROUTE_CRASH para indicarles que esta rama de la red ha caído
		agent_->send_qdrp_packet__route_crash();
	} 

	// cambia el estado a DESCONECTANDO
	agent_->change_state(qdrp_state::DISCONNECTING);	
}
// =============================================================================================================================================== //
// FIN DE: ESTADO = CONECTADO															   //
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = DESCONECTANDO															   //
// =============================================================================================================================================== //
// ======================================================================================================================== //
// Constructor / Destructor											     	    //
// ======================================================================================================================== //
qdrp_state__disconnecting::qdrp_state__disconnecting(QDRP* agent){ agent_=agent; }	// Constructor //

qdrp_state__disconnecting::~qdrp_state__disconnecting(){} // Destructor //


// ======================================================================================================================== //
// Inicialización de estado											     	    //
// ======================================================================================================================== //
void qdrp_state__disconnecting::initialize()
{
	fprintf(stdout, "[QDRP] at time %f, node %d is disconnecting\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	

	QDRP_global::cancel_timer(agent_->new_route_timer());

	// Comprueba si aún quedan conexiones por finalizar (con los nodos hijos o con el padre)
	// Cuando ha deshecho todos los enlaces, vuelve al estado OFFLINE.
	if(agent_->child_node_address_map()->is_empty() && agent_->parent_node_address()==(nsaddr_t) IP_BROADCAST)
	{
		agent_->change_state(qdrp_state::OFFLINE);
	} 	
}


// ======================================================================================================================== //
// Temporizadores												     	    //
// ======================================================================================================================== //
void qdrp_state__disconnecting::On_timer__join_network()
{
	fprintf(stdout, "[QDRP]: at time %f, node %d :: ERROR: received \"join_network\" timer signal while online\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);	
}

void qdrp_state__disconnecting::On_timer__heartbeat()
{
	fprintf(stdout, "[QDRP]: at time %f, node %d :: ERROR: received \"heartbeat\" timer signal while online\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);	
}


void qdrp_state__disconnecting::On_timer__connection_with_parent_node()
{
	switch(++agent_->parent_node_connection_timer()->cycle())
	{
		case 1:
		case 2:
		case 3:
		case 4:
		case 5: // espera a que expire el temporizador (100 % = 6 ciclos)
			// adicionalmente, se re-envían mensajes "desconectar" al nodo padre para acelerar el proceso, pues es posible que los
			// mensajes enviados con anterioridad se quedasen en el camino (si el medio es hostil, ¿qué se le va a hacer?)
			agent_->parent_node_connection_timer()->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
			agent_->send_qdrp_packet__disconnect(agent_->parent_node_address());
			break;
		case 6: /* TIMEOUT */
			agent_->parent_node_address() = IP_BROADCAST;
			// restablece el contador de ciclos del temporizador con el padre
			agent_->parent_node_connection_timer()->cycle()=0;
			// reinicia el estado DESCONECTADO
			agent_->change_state(qdrp_state::DISCONNECTING);
			break;
		default:
			fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: parent_node_connection_timer cycle counter went out of range\n",
				QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
			exit(-1);
			break;
	}
}
		
	

void qdrp_state__disconnecting::On_timer__connection_with_child_node(u_int8_t child_id)
{
	bool child_exists=false;
	// comprueba que el hijo con el identificador "child_id" exista
	nsaddr_t child_node_address = agent_->child_node_address_map()->get_address_by_identifier(child_id, &child_exists);
	
	if(!child_exists)
	{
		// esto no debe suceder
		fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received \"connection_with_child_node(%d)\" timer signal, associated with a node 				which does not exist\n", QDRP_global::CURRENT_TIME(), agent_->this_node()->address(), child_id);
		exit(-1);
	}

	switch(++agent_->child_node_connection_timer(child_id)->cycle())
	{
		case 1: 
		case 2:
		case 3:
		case 4:
		case 5: // espera a que expire el temporizador (100 % = 6 ciclos)
			// adicionalmente, se continúa enviando ROUTE_CRASH por si algún nodo hijo no se enteró de la pérdida de conexión de éste.
			agent_->child_node_connection_timer(child_id)->resched(QDRP_global::CONNECTION_TIMER_CYCLE_INTERVAL());
			agent_->send_qdrp_packet__route_crash();
			break;
		case 6: /* TIMEOUT */
			// des-registra el nodo hijo asociado al identificador
			agent_->child_node_address_map()->delete_entry(child_node_address);
			// restablece el contador de ciclos del temporizador
			agent_->child_node_connection_timer(child_id)->cycle()=0;
			// actualiza el nº de hijos del nodo
			agent_->this_node()->num_child_nodes()--;
			// vuelve al estado DESCONECTADO
			agent_->change_state(qdrp_state::DISCONNECTING);
			break;
		default:
			fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: parent_node_connection_timer cycle counter went out of range\n",
				QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
			exit(-1);
			break;
	}
}

void qdrp_state__disconnecting::On_timer__sendme_show_off()
{
}


void qdrp_state__disconnecting::On_timer__new_route()
{
	fprintf(stdout, "[QDRP]: at time %f, node %d :: ERROR: received \"new_route\" timer signal while disconnecting\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());
	exit(-1);	
}

// ======================================================================================================================== //
// Comandos													     	    //
// ======================================================================================================================== //
void qdrp_state__disconnecting::On_command__boot()
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}
void qdrp_state__disconnecting::On_command__boot_sink(u_int8_t network_id)
{
	fprintf(stdout, "[QDRP] at time %f :: INFO: node %d is already ON. \"boot\" command ignored\n",
		QDRP_global::CURRENT_TIME(), agent_->this_node()->address());	
}
void qdrp_state__disconnecting::On_command__switch_off()
{
	agent_->switch_off();
}



// ======================================================================================================================== //
// Recepción de paquetes											     	    //
// ======================================================================================================================== //
// paquete de datos //
void qdrp_state__disconnecting::On_reception_of_data_packet(Packet* p) 
{
	agent_->drop(p, DROP_RTR_NO_ROUTE);	// un nodo que está desconectando no acepta paquetes de datos
}


// paquetes QDRP //
void qdrp_state__disconnecting::On_reception_of_qdrp_packet__show_off(Packet* p){ Packet::free(p); }
void qdrp_state__disconnecting::On_reception_of_qdrp_packet__connect(Packet* p){ Packet::free(p); }
void qdrp_state__disconnecting::On_reception_of_qdrp_packet__connection_accepted(Packet* p){ Packet::free(p); }
void qdrp_state__disconnecting::On_reception_of_qdrp_packet__connection_rejected(Packet* p){ Packet::free(p); }
void qdrp_state__disconnecting::On_reception_of_qdrp_packet__connection_renew(Packet* p){ Packet::free(p); }
void qdrp_state__disconnecting::On_reception_of_qdrp_packet__route_crash(Packet* p){ Packet::free(p); }

void qdrp_state__disconnecting::On_reception_of_qdrp_packet__sendme_show_off(Packet* p){ Packet::free(p); }

void qdrp_state__disconnecting::On_reception_of_qdrp_packet__disconnect(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);

	// el mensaje debe estar destinado a este nodo
	if(ip_header->daddr() != agent_->this_node()->address()){ Packet::free(p); return; }

	// ¿el mensaje procede de un nodo hijo?
	bool child_exists=false;
	u_int8_t child_id=agent_->child_node_address_map()->get_identifier_by_address(ip_header->saddr(), &child_exists);
		
	if(child_exists)
	{
		// OK, el mensaje procede de un nodo hijo. Desactiva el temporizador asociado
		QDRP_global::cancel_timer(agent_->child_node_connection_timer(child_id));
		agent_->child_node_connection_timer(child_id)->cycle()=0;

		// Envía DESCONEXIÓN_OK al hijo
		agent_->send_qdrp_packet__disconnection_ok(ip_header->saddr());
	
		// des-registra el nodo hijo
		agent_->child_node_address_map()->delete_entry(ip_header->saddr());

		// y actualiza el nº de hijos del nodo
		agent_->this_node()->num_child_nodes()--;
	}
	Packet::free(p);

	// continúa en estado DESCONECTANDO ...
	agent_->change_state(qdrp_state::DISCONNECTING);
}


void qdrp_state__disconnecting::On_reception_of_qdrp_packet__disconnection_ok(Packet* p)
{
	// Obtiene la cabecera IP
	struct hdr_ip* ip_header=HDR_IP(p);

	// comprueba si el mensaje procede del nodo padre y que somos el destinatario
	if((agent_->parent_node_address() == ip_header->saddr()) && (agent_->this_node()->address() == ip_header->daddr()))
	{
		// desactiva el temporizador asociado al nodo padre
		QDRP_global::cancel_timer(agent_->parent_node_connection_timer());
		agent_->parent_node_connection_timer()->cycle()=0;

		// desregistra el nodo padre
		agent_->parent_node_address() = IP_BROADCAST;

		// Libera los recursos asociados al paquete
		Packet::free(p);
	
		// recarga el estado DESCONECTANDO 
		agent_->change_state(qdrp_state::DISCONNECTING);	
	}
	else Packet::free(p);
}



//================================================================================================================================ //
// FIN DE: ESTADO = DESCONECTANDO														   //
//================================================================================================================================ //





// **************************************************************************************************************************************************
// AGENTE QDRP
// **************************************************************************************************************************************************

/* Redefinición del método command() */
int QDRP::command(int argc, const char*const* argv)
{
	if (argc == 2)
	{
		if(strcasecmp(argv[1], "start") == 0)
		{
			return TCL_OK;
		}
		else if (strcasecmp(argv[1], "boot") == 0)
		{
			current_state()->On_command__boot();
			return TCL_OK;
		}
		else if (strcasecmp(argv[1], "switch-off") == 0)
		{
			current_state()->On_command__switch_off();
			return TCL_OK;
		}
		else if (strcasecmp(argv[1], "print_routing_info") == 0)
		{
			if (log_target_!=NULL)
			{
				sprintf(log_target_->pt_->buffer(), "P %f _%d_ node routing information", 
					QDRP_global::CURRENT_TIME(), this_node()->address());
				log_target_->pt_->dump();
				print_routing_info(log_target_);
			}
			else
			{
				fprintf(stdout, "%f _%d_ If you want to print this node's routing information, you must create a trace file in your 						Tcl simulation script", QDRP_global::CURRENT_TIME(), this_node()->address());
			}
			return TCL_OK;
		}
	}
	else if (argc == 3)
	{
		if (strcmp(argv[1], "boot-sink") == 0)
		{
			current_state()->On_command__boot_sink((u_int8_t) atoi(argv[2]));
			return TCL_OK;
		}
		else if (strcmp(argv[1], "port-dmux") == 0)
		{
			demultiplexer_ = (PortClassifier*) TclObject::lookup(argv[2]);
			if(demultiplexer_ == NULL)
			{
				fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1], argv[2]);
				return TCL_ERROR;
			}
			return TCL_OK;
		}
		else if (strcmp(argv[1], "log-target") == 0 || strcmp(argv[1], "tracetarget") == 0)
		{
			log_target_ = (Trace*) TclObject::lookup(argv[2]);
			if (log_target_ == NULL) { return TCL_ERROR; }
			return TCL_OK;
		}
	}
	
	return Agent::command(argc, argv);
}


/* Redefinición del método recv() */
void QDRP::recv(Packet* p, Handler* h)
{
	struct hdr_cmn* common_header = HDR_CMN(p);
	struct hdr_ip* ip_header = HDR_IP(p);

	// Decrementa el TTL (nº de saltos restantes) del paquete IP
	ip_header->ttl_--;
	if(ip_header->ttl_ == 0)
	{
		// oops...mal acabó tu carrera, cangurito.
		drop(p, DROP_RTR_TTL);
		return;
	}

	if (ip_header->saddr() == this_node()->address())	// Comprueba si el paquete está originado por este nodo
	{
		if(common_header->num_forwards() == 0)
		{
			common_header->size()+=IP_HDR_LEN;	// Añade el tamaño de la cabecera IP al tamaño del paquete
		}
		else if (common_header->num_forwards() > 0)
		{
			drop(p, DROP_RTR_ROUTE_LOOP);		// BUCLE detectado: elimina el paquete
			return;
		}
	}

	// En función de si el paquete entrante es de enrutamiento QDRP o de datos, llama al procedimiento adecuado
	if (common_header->ptype() == PT_QDRP)			
	{
		On_reception_of_qdrp_packet(p);
	}
	else
	{
		On_reception_of_data_packet(p);
	}
}

// **************************************************************************************************************************************************
// FIN DE: AGENTE QDRP
// **************************************************************************************************************************************************










// **************************************************************************************************************************************************
// CLASE QDRP (IMPLEMENTACIÓN)
// **************************************************************************************************************************************************
// ======================================================================================================================== //
// Constructor														    //
// ======================================================================================================================== //
QDRP::QDRP(nsaddr_t id) : Agent(PT_QDRP)
{
	// creación de los temporizadores
	join_network_timer_=new qdrp_timer__join_network(this);
	heartbeat_timer_=new qdrp_timer__heartbeat(this);
	parent_node_connection_timer_=new qdrp_timer__parent_node_connection(this);
	for(u_int8_t i=0; i < QDRP_global::SINK__NUM_CHILD_SLOTS; i++)
	{
		child_node_connection_timer_[i]=new qdrp_timer__child_node_connection(this, i);
	}

	sendme_show_off_=new qdrp_timer__sendme_show_off(this);
	new_route_=new qdrp_timer__new_route(this);

	// creación de la lista de nodos descubiertos
	discovered_node_list_ = new qdrp_parent_selection_list();

	// medio-inicialización del nodo local (se completa en el estado ARRANCANDO (BOOTING))
	parent_node_address() = (nsaddr_t) IP_BROADCAST;
	this_node() = new qdrp_node(id);
	child_node_address_map() = NULL;

	// inicialización de la máquina de estados
	for(u_int8_t i=0; i < qdrp_state::NUM_STATES; i++)
	{
		state_machine_[i]=qdrp_state::create_state(this, i);
	}
	change_state(qdrp_state::OFF);
}


// ======================================================================================================================== //
// Acceso a: Máquina de estados												    //
// ======================================================================================================================== //
// operación de cambio de estado //
void QDRP::change_state(u_int8_t next_state_id)
{
	assert(next_state_id < qdrp_state::NUM_STATES);

	current_state_id() = next_state_id;
	current_state()->initialize();
}


// ======================================================================================================================== //
// TXon y RXon de paquetes												    //
// ======================================================================================================================== //
Packet* QDRP::create_qdrp_packet(nsaddr_t destination_address)
{
	Packet* p=allocpkt();
	struct hdr_cmn* common_header=HDR_CMN(p);
	struct hdr_ip* ip_header=HDR_IP(p);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	// Parámetros de la cabecera común
	common_header->addr_type()=NS_AF_INET;
	common_header->direction()=hdr_cmn::DOWN;
	common_header->error()=0;
	common_header->next_hop()=destination_address;
	common_header->ptype()=PT_QDRP;
	common_header->size()=IP_HDR_LEN+QDRP_HEADER_LENGTH;
	common_header->xmit_failure_ = qdrp_mac_failed_callback;
	common_header->xmit_failure_data_ = (void*) this;
	
	// Parámetros de la cabecera IP
	ip_header->saddr()=this_node()->address();
	ip_header->daddr()=destination_address;
	ip_header->sport()=RT_PORT;
	ip_header->dport()=RT_PORT;
	ip_header->ttl()=IP_DEF_TTL;

	// Parámetros de la cabecera QDRP
	qdrp_header->message_type()=qdrp_packet::UNSPECIFIED;
	qdrp_header->creation_time()=QDRP_global::CURRENT_TIME();
	
	return p;
}



/* Evento de recepción de un paquete QDRP */
void QDRP::On_reception_of_qdrp_packet(Packet* p)
{
	struct hdr_qdrp* packet_header = HDR_QDRP(p); 		// Obtiene la cabecera QDRP
	struct hdr_ip* ip_header = HDR_IP(p);			// Obtiene la cabecera IP

	// El mensaje debe proceder de (y estar dirigido a) el puerto utilizado para enrutamiento, RT_PORT
	if((ip_header->sport() != RT_PORT) || (ip_header->dport() != RT_PORT))  { Packet::free(p); return; }

	switch(packet_header->message_type())
	{
		case qdrp_packet::SHOW_OFF:
			if(qdrp_packet::check_packet__show_off(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__show_off(p); 
			break;
		case qdrp_packet::CONNECT:
			if(qdrp_packet::check_packet__connect(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__connect(p); 
			break;
		case qdrp_packet::CONNECTION_ACCEPTED:
			if(qdrp_packet::check_packet__connection_accepted(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__connection_accepted(p); 
			break;
		case qdrp_packet::CONNECTION_REJECTED:
			if(qdrp_packet::check_packet__connection_rejected(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__connection_rejected(p); 
			break;
		case qdrp_packet::CONNECTION_RENEW:
			if(qdrp_packet::check_packet__connection_renew(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__connection_renew(p);
		case qdrp_packet::DISCONNECT:
			if(qdrp_packet::check_packet__disconnect(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__disconnect(p); 
			break;
		case qdrp_packet::DISCONNECTION_OK:
			if(qdrp_packet::check_packet__disconnection_ok(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__disconnection_ok(p); 
			break;
		case qdrp_packet::ROUTE_CRASH:
			if(qdrp_packet::check_packet__route_crash(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__route_crash(p); 
			break;
		case qdrp_packet::SENDME_SHOW_OFF:	//xendme
			if(qdrp_packet::check_packet__sendme_show_off(p)==false){ Packet::free(p); return; }
			current_state()->On_reception_of_qdrp_packet__sendme_show_off(p);
			break;
		default:
			Packet::free(p);  // Libera los recursos asociados al paquete
			fprintf(stdout, "[QDRP] at time %f, node %d :: ERROR: received unknown packet type: %d\n", 
				QDRP_global::CURRENT_TIME(), this_node()->address(), packet_header->message_type());
			exit(-1);	
			break;
	}
}


/* Evento de recepción de un paquete de datos */
void QDRP::On_reception_of_data_packet(Packet* p)
{
	// ¿Acaso esta función podría ser más fácil? :)
	current_state()->On_reception_of_data_packet(p);
}


/* Programa el envío de un paquete de datos */
void QDRP::send_data_packet(Packet* p)
{
	Scheduler::instance().schedule(target_, p, QDRP_global::DATA_PACKET_DELAY());
}


/* Programa el envío de un paquete QDRP */
void QDRP::send_qdrp_broadcast_packet(Packet* p)
{
	Scheduler::instance().schedule(target_, p, QDRP_global::QDRP_BROADCAST_PACKET_DELAY());
}

void QDRP::send_qdrp_unicast_packet(Packet* p)
{
	Scheduler::instance().schedule(target_, p, QDRP_global::QDRP_UNICAST_PACKET_DELAY());
}


/* Retransmite un paquete */
void QDRP::resend_packet(Packet* p)
{
	Scheduler::instance().schedule(target_, p, QDRP_global::PACKET_RETRANSMISSION_DELAY());
}


/* Envío de paquetes QDRP */
void QDRP::send_qdrp_packet__show_off()
{
	Packet* p=create_qdrp_packet((nsaddr_t) IP_BROADCAST);
	struct hdr_qdrp_packet__show_off* qdrp_header=HDR_QDRP_PACKET__SHOW_OFF(p);

	qdrp_header->message_type() = qdrp_packet::SHOW_OFF;
	qdrp_header->network_id() = this_node()->network_id();
	qdrp_header->hops_to_sink() = this_node()->hops_to_sink();
	qdrp_header->num_child_nodes() = this_node()->num_child_nodes();
	qdrp_header->max_child_nodes() = this_node()->max_child_nodes();

	send_qdrp_broadcast_packet(p);
}


void QDRP::send_qdrp_packet__connect(nsaddr_t destination)
{
	Packet* p=create_qdrp_packet(destination);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::CONNECT;

	send_qdrp_unicast_packet(p);
}


void QDRP::send_qdrp_packet__connection_accepted(nsaddr_t destination)
{
	Packet* p=create_qdrp_packet(destination);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::CONNECTION_ACCEPTED;

	send_qdrp_unicast_packet(p);
}


void QDRP::send_qdrp_packet__connection_rejected(nsaddr_t destination)
{
	Packet* p=create_qdrp_packet(destination);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::CONNECTION_REJECTED;

	send_qdrp_unicast_packet(p);
}


void QDRP::send_qdrp_packet__connection_renew(nsaddr_t destination)
{
	Packet *p=create_qdrp_packet(destination);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::CONNECTION_RENEW;

	send_qdrp_unicast_packet(p);
}


void QDRP::send_qdrp_packet__disconnect(nsaddr_t destination)
{
	Packet* p=create_qdrp_packet(destination);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::DISCONNECT;

	send_qdrp_unicast_packet(p);
}


void QDRP::send_qdrp_packet__disconnection_ok(nsaddr_t destination)
{
	Packet* p=create_qdrp_packet(destination);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::DISCONNECTION_OK;

	send_qdrp_unicast_packet(p);
}


void QDRP::send_qdrp_packet__route_crash()
{
	Packet* p=create_qdrp_packet((nsaddr_t) IP_BROADCAST);
	struct hdr_qdrp* qdrp_header=HDR_QDRP(p);

	qdrp_header->message_type()=qdrp_packet::ROUTE_CRASH;

	send_qdrp_broadcast_packet(p);
}

void QDRP::send_qdrp_packet__sendme_show_off() //xendme
{
	Packet* p=create_qdrp_packet((nsaddr_t) IP_BROADCAST);
	struct hdr_qdrp_packet__sendme_show_off* qdrp_header=HDR_QDRP_PACKET__SENDME_SHOW_OFF(p);

	qdrp_header->message_type() = qdrp_packet::SENDME_SHOW_OFF;
	qdrp_header->network_id() = this_node()->network_id();

	send_qdrp_broadcast_packet(p);
}

// ======================================================================================================================== //
// Operaciones														    //
// ======================================================================================================================== //
void QDRP::print_routing_info(Trace* out)
{
	sprintf(out->pt_->buffer(), "NODE %d ROUTING INFORMATION", this_node()->address());
	out->pt_->dump();
	sprintf(out->pt_->buffer(), "   - PARENT NODE: %d", parent_node_address());
	out->pt_->dump();
	if(child_node_address_map()!=NULL)
	{
		child_node_address_map()->dump_on_trace(out);
	}
	sprintf(out->pt_->buffer(), "\n");
	out->pt_->dump();
}


void QDRP::switch_off()
{
	/* ====================================================================== */
	/*                         LIMPIEZA GENERAL                               */
	/* ====================================================================== */

	// Vacía la lista de nodos descubiertos
	discovered_node_list()->clear();

	// Borra del mapa al mapa (...¿?) de direcciones de nodos hijos
	if(child_node_address_map()!=NULL)
	{
		delete child_node_address_map();
		child_node_address_map()=NULL;
	}
	// Establece como nodo padre a Don Nadie
	parent_node_address() = IP_BROADCAST;
	
	// Fija los parámetros propios de un nodo desconectado
	this_node()->network_id() = QDRP_global::DEFAULT_NETWORK_ID;
	this_node()->hops_to_sink() = QDRP_global::UNKNOWN_HOPS_TO_SINK;
	this_node()->num_child_nodes() = 0;
	this_node()->max_child_nodes() = QDRP_global::NODE__NUM_CHILD_SLOTS;

	// Cancela los temporizadores pendientes
	QDRP_global::cancel_timer(join_network_timer());
	QDRP_global::cancel_timer(heartbeat_timer());


	QDRP_global::cancel_timer(sendme_show_off_timer());
	QDRP_global::cancel_timer(new_route_timer());

	QDRP_global::cancel_timer(parent_node_connection_timer()); parent_node_connection_timer()->cycle()=0;
	for(int i=0; i<=0xFF; i++)
	{
		qdrp_timer__child_node_connection* timer=child_node_connection_timer((u_int8_t) i);
		if(timer==NULL) break;
		else 
		{
			QDRP_global::cancel_timer(child_node_connection_timer((u_int8_t) i));
			child_node_connection_timer((u_int8_t) i)->cycle()=0;
		}
	}

	change_state(qdrp_state::OFF);
}


// ======================================================================================================================== //
// Callbacks MAC													    //
// ======================================================================================================================== //
void QDRP::On_mac_failure(Packet* p)
{
	struct hdr_ip* ip_header=HDR_IP(p);
	struct hdr_cmn* common_header=HDR_CMN(p);

	// El que sea de datos o de enrutamiento es irrelevante en este caso. Envía este paquete usando fuerza bruta.
	resend_packet(p); 
}

void QDRP::qdrp_mac_failed_callback(Packet* p, void* arg)
{
	((QDRP*) arg)->On_mac_failure(p);
}
	
// ****************************************************************************************************************************
// FIN DE CLASE QDRP (IMPLEMENTACIÓN)
// ****************************************************************************************************************************

