#ifndef __qdrp_address_map_h__
#define __qdrp_address_map_h__

#include <qdrp/qdrp_imports.h>
#include <qdrp/qdrp_global.h>

// **************************************************************************************************************************************************
// MAPA DE DIRECCIONES
// **************************************************************************************************************************************************

class qdrp_address_map
{
	//==================================================================================================================================//
	// Características de un mapa de direcciones											    //
	//==================================================================================================================================//
	std::map<nsaddr_t, u_int8_t> direct_seek_table_;		// tabla de búsqueda directa (dirección IP => identificador)
	std::map<u_int8_t, nsaddr_t> reverse_seek_table_;		// tabla de búsqueda inversa (identificador => dirección IP)
	std::deque<u_int8_t> identifier_queue_;				// cola donde se almacenan los identificadores no usados
	u_int8_t max_size_;						// tamaño máximo (nº de entradas) del mapa de direcciones
public:
	//==================================================================================================================================//
	// Constructor / Destructor													    //
	//==================================================================================================================================//
	qdrp_address_map(u_int8_t max_size);				// constructor
	~qdrp_address_map();						// destructor
	void clear();							// limpieza / vaciado / borrado / bla bla bla

	//==================================================================================================================================//
	// Funciones de consulta													    //
	//==================================================================================================================================//
	bool is_empty();						// devuelve "true" si el mapa está vacío, "false" en caso contrario
	bool is_full();							//    "       "    "  "    "    "  lleno,     "    "   "      "
	u_int8_t get_identifier_by_address(nsaddr_t address, bool* found);	// obtiene el identificador asociado a una dirección IP
	nsaddr_t get_address_by_identifier(u_int8_t id, bool* found);		// obtiene la dirección IP asociada a un identificador
	u_int8_t size();							// tamaño actual (nº de entradas) del mapa de direcciones
	u_int8_t max_size();							// tamaño máximo del mapa de direcciones
	void dump_on_trace(Trace* out);						// Muestra el contenido del mapa de direcciones

	//==================================================================================================================================//
	// Modificadores														    //
	//==================================================================================================================================//
	void add_entry(nsaddr_t address);					// añade una nueva entrada al mapa
	void delete_entry(nsaddr_t address);					// borra una entrada del mapa, dada su dirección
	void delete_entry_by_id(u_int8_t identifier);				//   "    "     "     "   "    dado su identificador
};

// **************************************************************************************************************************************************
// FIN DE: MAPA DE DIRECCIONES
// **************************************************************************************************************************************************

#endif

