 #include <qdrp/qdrp_packet.h>

// =============================================================================================================================================== //
// TIPOS DE PAQUETE QDRP															   //
// =============================================================================================================================================== //

//==========================================================================================================================//
// Funciones de comprobación de validez										     	    //
//==========================================================================================================================//

bool qdrp_packet::check_packet__show_off(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);
	struct hdr_qdrp_packet__show_off* qdrp_header = HDR_QDRP_PACKET__SHOW_OFF(p);

	// La dirección origen de un SHOW_OFF debe pertenecer a un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// Los SHOW_OFFs son mensajes de DIFUSIÓN
	if(ip_header->daddr() != (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// ¡Tiene que tratarse de un paquete SHOW_OFF!
	if(qdrp_header->message_type() != qdrp_packet::SHOW_OFF){ return false; }

	// El número de saltos a un sumidero debe ser menor o igual que el máximo permitido
	if(qdrp_header->hops_to_sink() > QDRP_global::MAX_HOPS_TO_SINK){ return false; }

	// El número de nodos hijos conectados actualmente a un nodo debe ser igual o inferior al nº máximo de nodos soportados
	if(qdrp_header->num_child_nodes() > qdrp_header->max_child_nodes()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// OK...aceptamos barco.
	return true;
}


bool qdrp_packet::check_packet__connect(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);
	struct hdr_qdrp_packet__connect* qdrp_header = HDR_QDRP_PACKET__CONNECT(p);

	// Asegurarse de que se trata de un mensaje CONNECT
	if(qdrp_header->message_type() != qdrp_packet::CONNECT){ return false; }

	// La dirección origen de un CONNECT debe pertenecer a un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje CONNECT es unicast, por tanto la dirección destino es la de un nodo
	if(ip_header->daddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// Bien, el formato del mensaje es correcto.
	return true;
}


bool qdrp_packet::check_packet__connection_accepted(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);						
	struct hdr_qdrp_packet__connection_accepted* qdrp_header = HDR_QDRP_PACKET__CONNECTION_ACCEPTED(p);	

	// mensaje CONNECTION_ACCEPTED
	if(qdrp_header->message_type() != qdrp_packet::CONNECTION_ACCEPTED){ return false; }

	// La dirección origen de un CONNECTION_ACCEPTED debe ser la de un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje CONNECTION_ACCEPTED es unicast
	if(ip_header->daddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// Correcto.
	return true;
}


bool qdrp_packet::check_packet__connection_rejected(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);						
	struct hdr_qdrp_packet__connection_rejected* qdrp_header = HDR_QDRP_PACKET__CONNECTION_REJECTED(p);	

	// mensaje CONNECTION_REJECTED
	if(qdrp_header->message_type() != qdrp_packet::CONNECTION_REJECTED){ return false; }

	// La dirección origen de un CONNECTION_REJECTED debe ser la de un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje CONNECTION_REJECTED es unicast
	if(ip_header->daddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// Correcto.
	return true;
}


bool qdrp_packet::check_packet__connection_renew(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);						
	struct hdr_qdrp_packet__connection_renew* qdrp_header = HDR_QDRP_PACKET__CONNECTION_RENEW(p);	

	// mensaje CONNECTION_RENEW
	if(qdrp_header->message_type() != qdrp_packet::CONNECTION_RENEW){ return false; }

	// La dirección origen de un CONNECTION_RENEW debe ser la de un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje CONNECTION_RENEW es unicast
	if(ip_header->daddr() == (nsaddr_t) IP_BROADCAST) { return false; }
	
	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }
	
	// Correcto.
	return true;
}


bool qdrp_packet::check_packet__disconnect(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);						
	struct hdr_qdrp_packet__disconnect* qdrp_header = HDR_QDRP_PACKET__DISCONNECT(p);	

	// mensaje DISCONNECT
	if(qdrp_header->message_type() != qdrp_packet::DISCONNECT){ return false; }

	// La dirección origen de un DISCONNECT debe ser la de un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje DISCONNECT es unicast
	if(ip_header->daddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// Correcto.
	return true;
}


bool qdrp_packet::check_packet__disconnection_ok(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);						
	struct hdr_qdrp_packet__disconnection_ok* qdrp_header = HDR_QDRP_PACKET__DISCONNECTION_OK(p);	

	// mensaje DISCONNECTION_OK
	if(qdrp_header->message_type() != qdrp_packet::DISCONNECTION_OK){ return false; }

	// La dirección origen de un DISCONNECTION_OK debe ser la de un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje DISCONNECTION_OK es unicast
	if(ip_header->daddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// Correcto.
	return true;
}


bool qdrp_packet::check_packet__route_crash(Packet* p)
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);						
	struct hdr_qdrp_packet__route_crash* qdrp_header = HDR_QDRP_PACKET__ROUTE_CRASH(p);	

	// mensaje ROUTE_CRASH
	if(qdrp_header->message_type() != qdrp_packet::ROUTE_CRASH){ return false; }

	// La dirección origen de un ROUTE_CRASH debe ser la de un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// El mensaje ROUTE_CRASH es de difusión
	if(ip_header->daddr() != (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// Correcto.
	return true;

}

bool qdrp_packet::check_packet__sendme_show_off(Packet* p)	//xendme
{
	// Obtiene las cabeceras IP y QDRP
	struct hdr_ip* ip_header = HDR_IP(p);
	struct hdr_qdrp_packet__show_off* qdrp_header = HDR_QDRP_PACKET__SHOW_OFF(p);

	// La dirección origen de un SENDME SHOW_OFF debe pertenecer a un nodo
	if(ip_header->saddr() == (nsaddr_t) IP_BROADCAST) { return false; }

	// Los SENDME SHOW_OFFs son mensajes de DIFUSIÓN
	if(ip_header->daddr() != (nsaddr_t) IP_BROADCAST) { return false; }

	// La dirección origen debe ser distinta a la dirección destino
	if(ip_header->saddr() == ip_header->daddr()){ return false; }

	// ¡Tiene que tratarse de un paquete SENDME_SHOW_OFF!
	if(qdrp_header->message_type() != qdrp_packet::SENDME_SHOW_OFF){ return false; }


	// Comprueba que el instante de creación del mensaje sea anterior al instante actual...¡pero no demasiado!
	if(qdrp_header->creation_time() > QDRP_global::CURRENT_TIME() ||
		qdrp_header->creation_time() < QDRP_global::CURRENT_TIME()-QDRP_global::CONNECTION_TIMEOUT()){ return false; }

	// OK...aceptamos barco.
	return true;
}

// =============================================================================================================================================== //
// FIN DE: TIPOS DE PAQUETE QDRP														   //
// =============================================================================================================================================== //

