#ifndef __qdrp_packet_h__
#define __qdrp_packet_h__

#include <qdrp/qdrp_imports.h>
#include <qdrp/qdrp_global.h>

// =============================================================================================================================================== //
// TIPOS DE PAQUETE QDRP															   //
// =============================================================================================================================================== //

class qdrp_packet
{
public:
	//==========================================================================================================================//
	// Tipos de paquete	 											     	    //
	//==========================================================================================================================//
	static const u_int8_t UNSPECIFIED = 0x00;
	static const u_int8_t SHOW_OFF = 0x01;
	static const u_int8_t CONNECT = 0x02;
	static const u_int8_t CONNECTION_ACCEPTED = 0x03;
	static const u_int8_t CONNECTION_REJECTED = 0x04;
	static const u_int8_t CONNECTION_RENEW = 0x05;
	static const u_int8_t DISCONNECT = 0x06;
	static const u_int8_t DISCONNECTION_OK = 0x07;
	static const u_int8_t ROUTE_CRASH = 0x08;
	static const u_int8_t SENDME_SHOW_OFF = 0x09;	//xendme

	//==========================================================================================================================//
	// Funciones de comprobación de validez										     	    //
	//==========================================================================================================================//
	static bool check_packet__show_off(Packet* p);
	static bool check_packet__connect(Packet* p);
	static bool check_packet__connection_accepted(Packet* p);
	static bool check_packet__connection_rejected(Packet* p);
	static bool check_packet__connection_renew(Packet* p);
	static bool check_packet__disconnect(Packet* p);
	static bool check_packet__disconnection_ok(Packet* p);
	static bool check_packet__route_crash(Packet* p);

	static bool check_packet__sendme_show_off(Packet* p);	//xendme
};

// =============================================================================================================================================== //
// FIN DE: TIPOS DE PAQUETE QDRP														   //
// =============================================================================================================================================== //





// ================================================================================================================================================ //
// CONSTANTES
// ================================================================================================================================================ //

#define QDRP_HEADER_LENGTH 	sizeof(hdr_all_qdrp)


// **************************************************************************************************************************************************
// CABECERA DE LOS PAQUETES QDRP
// **************************************************************************************************************************************************

#define HDR_QDRP(p) 				((struct hdr_qdrp*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__SHOW_OFF(p)		((struct hdr_qdrp_packet__show_off*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__CONNECT(p)		((struct hdr_qdrp_packet__connect*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__CONNECTION_ACCEPTED(p)	((struct hdr_qdrp_packet__connection_accepted*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__CONNECTION_REJECTED(p)	((struct hdr_qdrp_packet__connection_rejected*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__CONNECTION_RENEW(p)	((struct hdr_qdrp_packet__connection_renew*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__DISCONNECT(p) 		((struct hdr_qdrp_packet__disconnect*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__DISCONNECTION_OK(p)	((struct hdr_qdrp_packet__disconnection_ok*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__ROUTE_CRASH(p)		((struct hdr_qdrp_packet__route_crash*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__SLEEP(p) 		((struct hdr_qdrp_packet__sleep*) hdr_qdrp::access(p))
#define HDR_QDRP_PACKET__SENDME_SHOW_OFF(p)	((struct hdr_qdrp_packet__sendme_show_off*) hdr_qdrp::access(p))

struct hdr_qdrp
{
	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje

	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }

	static int offset_;
	inline static int& offset(){ return offset_; }
	inline static hdr_qdrp* access(const Packet* p){ return (hdr_qdrp*) p->access(offset_); }
};

struct hdr_qdrp_packet__show_off
{
	// cabecera_IP.origen = < dirección IP del nodo local >
	// cabecera_IP.destino = IP_BROADCAST

	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje
	u_int8_t network_id_;			// identificador de la red a la que el nodo está conectado
	u_int8_t hops_to_sink_;			// nº de saltos hasta el sumidero de la red
	u_int8_t num_child_nodes_;		// nº de nodos hijos conectados al nodo
	u_int8_t max_child_nodes_;		// nº maximo de nodos hijos que puede soportar el nodo
	
	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
	inline u_int8_t& network_id(){ return network_id_; }
	inline u_int8_t& hops_to_sink(){ return hops_to_sink_; }
	inline u_int8_t& num_child_nodes(){ return num_child_nodes_; }
	inline u_int8_t& max_child_nodes(){ return max_child_nodes_; }
};

struct hdr_qdrp_packet__connect
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = < dirección del nodo padre >

	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje
	
	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};

struct hdr_qdrp_packet__connection_accepted
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = < dirección del nodo "candidato a hijo"

	double creation_time_;			// instante de creación 
	u_int8_t message_type_;			// tipo de mensaje
	
	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};

struct hdr_qdrp_packet__connection_rejected
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = < dirección del nodo "candidato a hijo"

	double creation_time_;			// instante de creación 
	u_int8_t message_type_;			// tipo de mensaje

	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};

struct hdr_qdrp_packet__connection_renew
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = < dirección del nodo hijo >

	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje

	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};


struct hdr_qdrp_packet__disconnect
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = < dirección IP del nodo padre >

	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje

	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};

struct hdr_qdrp_packet__disconnection_ok
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = < dirección IP del nodo hijo >

	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje

	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};

struct hdr_qdrp_packet__route_crash
{
	// cabecera_IP.origen = < dirección IP del nodo local > 
	// cabecera_IP.destino = IP_BROADCAST	
	
	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje

	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
};

struct hdr_qdrp_packet__sendme_show_off		//xendme
{
	// cabecera_IP.origen = < dirección IP del nodo local >
	// cabecera_IP.destino = IP_BROADCAST

	double creation_time_;			// instante de creación
	u_int8_t message_type_;			// tipo de mensaje
	u_int8_t network_id_;			// identificador de la red a la que el nodo está conectado
	
	inline double& creation_time(){ return creation_time_; }
	inline u_int8_t& message_type(){ return message_type_; }
	inline u_int8_t& network_id(){ return network_id_; }
};

// para calcular la reserva de espacio de cabecera
union hdr_all_qdrp
{
	hdr_qdrp				base_;
	hdr_qdrp_packet__show_off		show_off_;
	hdr_qdrp_packet__connect		connect_;
	hdr_qdrp_packet__connection_accepted	connection_accepted_;
	hdr_qdrp_packet__connection_rejected	connection_rejected_;
	hdr_qdrp_packet__connection_renew	connection_renew_;
	hdr_qdrp_packet__disconnect		disconnect_;
	hdr_qdrp_packet__disconnection_ok	disconnection_ok_;
	hdr_qdrp_packet__route_crash		route_crash_;
        hdr_qdrp_packet__sendme_show_off	sendme_show_off_; //xendme
};

// **************************************************************************************************************************************************
// FIN DE: CABECERA DE LOS PAQUETES QDRP
// **************************************************************************************************************************************************

#endif

