#include <qdrp/qdrp_parent_selection_list.h>


// *********************************************************************************************************************************************** //
// LISTA DE SELECCIÓN DE CANDIDATOS A NODO PADRE
// *********************************************************************************************************************************************** //

//==========================================================================================================================//
// Constructor/Destructor 											     	    //
//==========================================================================================================================//
qdrp_parent_selection_list::qdrp_parent_selection_list(){ clear(); }			// Constructor

qdrp_parent_selection_list::~qdrp_parent_selection_list(){}				// Destructor

void qdrp_parent_selection_list::clear(){ node_list_.clear(); is_sorted_ = true; } 	// Borrado

//==========================================================================================================================//
// Métodos de consulta 												     	    //
//==========================================================================================================================//
bool qdrp_parent_selection_list::is_empty(){ return node_list_.empty(); }	// comprueba si la lista está vacía

/* Obtiene el mejor candidato a nodo padre registrado hasta el momento	*/
qdrp_node qdrp_parent_selection_list::get_best_candidate()
{
	assert(!is_empty());
//
//
//	is_sorted=false;	////diego - aqui  -  xendme

	if(!is_sorted_)
	{
		node_list_.sort(qdrp_node::compare);
		is_sorted_ = true;
	}

	return node_list_.front();
}	


//==========================================================================================================================//
// Modificadores												     	    //
//==========================================================================================================================//
/* Añade, actualiza o descarta un nodo en la lista */
void qdrp_parent_selection_list::register_node(qdrp_node entry)
{
	bool is_useful_entry = entry.num_child_nodes() < entry.max_child_nodes();
	// en primer lugar, busca si ya existía en la lista una entrada con la misma dirección que la nueva
	std::list<qdrp_node>::iterator it=find_node_by_address(entry.address());
	if (it!=node_list_.end())
	{
		// ha encontrado una entrada de ese mismo nodo
		if(is_useful_entry)
		{
			// actualiza la entrada
			it->network_id() = entry.network_id();
			it->hops_to_sink() = entry.hops_to_sink();
			it->num_child_nodes() = entry.num_child_nodes();
			it->max_child_nodes() = entry.max_child_nodes();
//			it->route_batt()= entry.route_batt();
			is_sorted_ = false; // cuando se modifica algún dato en la lista, no se garantiza que ésta esté ordenada.
		}
		else
		{
			// esta entrada ya no es útil en la lista de nodos candidatos, p.ej: el nodo candidato no puede aceptar más conexiones.
			node_list_.erase(it);
		}
	}
	else
	{
		// entrada de este nodo no encontrada en la lista
		if(is_useful_entry)
		{
			// la información es útil. Al saco.
			node_list_.push_back(entry);
			is_sorted_ = false; // Al añadir una nueva entrada en la lista, tampoco se garantiza que esté ordenada
		}
	}
}


/* Elimina la primera entrada de la lista (mejor candidato a nodo padre) */
void qdrp_parent_selection_list::delete_best_candidate()
{
	assert(!is_empty());		// No se puede eliminar donde no hay.
	node_list_.pop_front();
}
 

//==========================================================================================================================//
// Funciones privadas 												     	    //
//==========================================================================================================================//
/* busca un nodo en la lista, dada su dirección IP. */
std::list<qdrp_node>::iterator qdrp_parent_selection_list::find_node_by_address(nsaddr_t address)
{
	std::list<qdrp_node>::iterator it;
	for(it=node_list_.begin() ; it != node_list_.end(); it++)
	{
		if(address==it->address()) break;
	}
	return it;
}


// *********************************************************************************************************************************************** //
// FIN DE: LISTA DE SELECCIÓN DE CANDIDATOS A NODO PADRE
// *********************************************************************************************************************************************** //

